@@ -2,9 +2,16 @@ import itertools
 import sys
 from collections import defaultdict
 import random
+import time
 import warnings
 from xml.sax.saxutils import escape
 
+try:
+    import dask
+    import dask.array as da
+except ImportError:
+    dask = None
+
 from AnyQt.QtWidgets import QWidget, QGraphicsItem, QPushButton, QMenu, \
     QGridLayout, QAction, QVBoxLayout, QApplication, QWidgetAction, \
     QShortcut, QToolTip, QGraphicsRectItem, QGraphicsTextItem
@@ -36,6 +43,7 @@ from Orange.widgets.utils.concurrent import TaskState, ConcurrentMixin
 from Orange.widgets.visualize.utils.plotutils import HelpEventDelegate, PlotWidget
 from Orange.widgets.visualize.utils.customizableplot import CommonParameterSetter
 
+from orangecontrib.spectroscopy import dask_client
 from orangecontrib.spectroscopy.data import getx
 from orangecontrib.spectroscopy.utils import apply_columns_numpy
 from orangecontrib.spectroscopy.widgets.line_geometry import \
@@ -272,32 +280,34 @@ class InterruptException(Exception):
 
 class ShowAverage(QObject, ConcurrentMixin):
 
-    average_shown = pyqtSignal()
+    shown = pyqtSignal()
 
     def __init__(self, master):
         super().__init__(parent=master)
         ConcurrentMixin.__init__(self)
         self.master = master
 
-    def show_average(self):
+    def show(self):
         master = self.master
         master.clear_graph()  # calls cancel
         master.view_average_menu.setChecked(True)
         master.set_pen_colors()
         master.viewtype = AVERAGE
         if not master.data:
-            self.average_shown.emit()
+            self.shown.emit()
         else:
             color_var = master.feature_color
             self.start(self.compute_averages, master.data, color_var, master.subset_indices,
-                       master.selection_group, master.data_xsind, master.selection_type)
+                       master.selection_group, master.selection_type)
 
     @staticmethod
     def compute_averages(data: Orange.data.Table, color_var, subset_indices,
-                         selection_group, data_xsind, selection_type, state: TaskState):
+                         selection_group, selection_type, state: TaskState):
 
         def progress_interrupt(i: float):
             if state.is_interruption_requested():
+                if future:
+                    future.cancel()
                 raise InterruptException
 
         def _split_by_color_value(data, color_var):
@@ -318,7 +328,12 @@ class ShowAverage(QObject, ConcurrentMixin):
 
         results = []
 
+        future = None
+
+        is_dask = dask and isinstance(data.X, dask.array.Array)
+
         dsplit = _split_by_color_value(data, color_var)
+        compute_dask = []
         for colorv, indices in dsplit.items():
             for part in [None, "subset", "selection"]:
                 progress_interrupt(0)
@@ -329,17 +344,34 @@ class ShowAverage(QObject, ConcurrentMixin):
                 elif part == "subset":
                     part_selection = indices & subset_indices
                 if np.any(part_selection):
-                    std = apply_columns_numpy(data.X,
-                                              lambda x: bottleneck.nanstd(x, axis=0),
-                                              part_selection,
-                                              callback=progress_interrupt)
-                    mean = apply_columns_numpy(data.X,
-                                               lambda x: bottleneck.nanmean(x, axis=0),
-                                               part_selection,
-                                               callback=progress_interrupt)
-                    std = std[data_xsind]
-                    mean = mean[data_xsind]
-                    results.append((colorv, part, mean, std, part_selection))
+                    if is_dask:
+                        subset = data.X[part_selection]
+                        compute_dask.extend([da.nanstd(subset, axis=0),
+                                             da.nanmean(subset, axis=0)])
+                        std, mean = None, None
+                    else:
+                        std = apply_columns_numpy(data.X,
+                                                  lambda x: bottleneck.nanstd(x, axis=0),
+                                                  part_selection,
+                                                  callback=progress_interrupt)
+                        mean = apply_columns_numpy(data.X,
+                                                   lambda x: bottleneck.nanmean(x, axis=0),
+                                                   part_selection,
+                                                   callback=progress_interrupt)
+                    results.append([colorv, part, mean, std, part_selection])
+
+        if is_dask:
+            future = dask_client.compute(dask.array.vstack(compute_dask))
+            while not future.done():
+                progress_interrupt(0)
+                time.sleep(0.1)
+            if future.cancelled():
+                return
+            computed = future.result()
+            for i, lr in enumerate(results):
+                lr[2] = computed[i*2]
+                lr[3] = computed[i*2+1]
+
         progress_interrupt(0)
         return results
 
@@ -352,6 +384,8 @@ class ShowAverage(QObject, ConcurrentMixin):
         x = master.data_x
 
         for colorv, part, mean, std, part_selection in res:
+            std = std[master.data_xsind]
+            mean = mean[master.data_xsind]
             if part is None:
                 pen = master.pen_normal if np.any(master.subset_indices) else master.pen_subset
             elif part == "selection" and master.selection_type:
@@ -370,16 +404,106 @@ class ShowAverage(QObject, ConcurrentMixin):
         master.curves_cont.update()
         master.plot.vb.set_mode_panning()
 
-        self.average_shown.emit()
+        self.shown.emit()
 
     def on_partial_result(self, result):
         pass
 
-    def on_exception(self, ex: Exception):
-        if isinstance(ex, InterruptException):
-            return
 
-        raise ex
+class ShowIndividual(QObject, ConcurrentMixin):
+
+    shown = pyqtSignal()
+
+    def __init__(self, master):
+        super().__init__(parent=master)
+        ConcurrentMixin.__init__(self)
+        self.master = master
+
+    def show(self):
+        master = self.master
+        master.clear_graph()  # calls cancel
+        master.view_average_menu.setChecked(False)
+        master.set_pen_colors()
+        master.viewtype = INDIVIDUAL
+        if not master.data:
+            return
+        sampled_indices = master._compute_sample(master.data.X)
+        self.start(self.compute_curves, master.data_x, master.data.X,
+                   sampled_indices)
+
+    @staticmethod
+    def compute_curves(x, ys, sampled_indices, state: TaskState):
+        is_dask = dask and isinstance(ys, dask.array.Array)
+
+        def progress_interrupt(i: float):
+            if state.is_interruption_requested():
+                if future:
+                    future.cancel()
+                raise InterruptException
+
+        future = None
+
+        progress_interrupt(0)
+        ys = ys[sampled_indices]
+        if is_dask:
+            future = dask_client.compute(ys)
+            while not future.done():
+                progress_interrupt(0)
+                time.sleep(0.1)
+            if future.cancelled():
+                return
+            ys = future.result()
+        ys[np.isinf(ys)] = np.nan  # remove infs that could ruin display
+
+        progress_interrupt(0)
+        return x, ys, sampled_indices
+
+    def on_done(self, res):
+        x, ys, sampled_indices = res
+
+        master = self.master
+        ys = ys[:, master.data_xsind]
+
+        if master.waterfall:
+            waterfall_constant = 0.1
+            miny = bottleneck.nanmin(ys)
+            maxy = bottleneck.nanmax(ys)
+            space = (maxy - miny) * waterfall_constant
+            mul = (np.arange(len(ys))*space + 1).reshape(-1, 1)
+            ys = ys * mul
+
+        # shuffle the data before drawing because classes often appear sequentially
+        # and the last class would then seem the most prevalent if colored
+        indices = list(range(len(sampled_indices)))
+        random.Random(master.sample_seed).shuffle(indices)
+        sampled_indices = [sampled_indices[i] for i in indices]
+        master.sampled_indices = sampled_indices
+        master.sampled_indices_inverse = {s: i for i, s in enumerate(master.sampled_indices)}
+        master.new_sampling.emit(len(master.sampled_indices))
+        ys = ys[indices]  # ys was already subsampled
+
+        master.curves.append((x, ys))
+
+        # add curves efficiently
+        for y in ys:
+            master.add_curve(x, y, ignore_bounds=True)
+
+        if x.size and ys.size:
+            bounding_rect = QGraphicsRectItem(QRectF(
+                QPointF(bottleneck.nanmin(x), bottleneck.nanmin(ys)),
+                QPointF(bottleneck.nanmax(x), bottleneck.nanmax(ys))))
+            bounding_rect.setPen(QPen(Qt.NoPen))  # prevents border of 1
+            master.curves_cont.add_bounds(bounding_rect)
+
+        master.curves_plotted.append((x, ys))
+        master.set_curve_pens()
+        master.curves_cont.update()
+        master.plot.vb.set_mode_panning()
+
+        self.shown.emit()
+
+    def on_partial_result(self, result):
+        pass
 
 
 class InteractiveViewBox(ViewBox):
@@ -768,7 +892,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
 
     sample_seed = Setting(0, schema_only=True)
     peak_labels_saved = Setting([], schema_only=True)
-    feature_color = ContextSetting(None)
+    feature_color = ContextSetting(None, exclude_attributes=True)
     color_individual = Setting(False)  # color individual curves (in a cycle) if no feature_color
     invertX = Setting(False)
     viewtype = Setting(INDIVIDUAL)
@@ -779,6 +903,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
     new_sampling = pyqtSignal(object)
     highlight_changed = pyqtSignal()
     locked_axes_changed = pyqtSignal(bool)
+    graph_shown = pyqtSignal()
 
     def __init__(self, parent: OWWidget, select=SELECTNONE):
         QWidget.__init__(self)
@@ -786,7 +911,12 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
         SelectionGroupMixin.__init__(self)
 
         self.show_average_thread = ShowAverage(self)
-        self.show_average_thread.average_shown.connect(self.rescale)
+        self.show_average_thread.shown.connect(self.rescale)
+        self.show_average_thread.shown.connect(self.graph_shown.emit)
+
+        self.show_individual_thread = ShowIndividual(self)
+        self.show_individual_thread.shown.connect(self.rescale)
+        self.show_individual_thread.shown.connect(self.graph_shown.emit)
 
         self.parent = parent
 
@@ -1032,7 +1162,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
                 index = self.sampled_indices[self.highlighted]
                 variables = self.data.domain.metas + self.data.domain.class_vars
                 text += "".join(
-                    '{} = {}\n'.format(attr.name, self.data[index][attr])
+                    '{} = {}\n'.format(attr.name, self.data[index, attr])
                     for attr in variables)
             elif self.viewtype == AVERAGE:
                 c = self.multiple_curves_info[self.highlighted]
@@ -1136,6 +1266,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
 
     def clear_graph(self):
         self.show_average_thread.cancel()
+        self.show_individual_thread.cancel()
         self.highlighted = None
         # reset caching. if not, it is not cleared when view changing when zoomed
         self.curves_cont.setCacheMode(QGraphicsItem.NoCache)
@@ -1159,7 +1290,6 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
 
         self.sampled_indices = []
         self.sampled_indices_inverse = {}
-        self.sampling = None
         self.new_sampling.emit(None)
 
         for m in self.markings:
@@ -1192,6 +1322,8 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
 
         old_sel_ci = current_selection()
 
+        changed = False
+
         if add_to_group:  # both keys - need to test it before add_group
             selnum = np.max(self.selection_group)
         elif add_group:
@@ -1201,15 +1333,20 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
         else:
             # remove the current selection
             redraw_curve_indices.update(old_sel_ci)
-            self.selection_group *= 0  # remove
+            if np.any(self.selection_group):
+                self.selection_group *= 0  # remove
+                changed = True
             selnum = 1
         # add new
         if data_indices is not None:
             self.selection_group[data_indices] = selnum
             redraw_curve_indices.update(
                 icurve for idata, icurve in invd.items() if idata in data_indices_set)
+            changed = True
 
         fixes = self.make_selection_valid()
+        if fixes:
+            changed = True
         redraw_curve_indices.update(
             icurve for idata, icurve in invd.items() if idata in fixes)
 
@@ -1221,7 +1358,8 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
             redraw_curve_indices.update(old_sel_ci)
 
         self.set_curve_pens(redraw_curve_indices)
-        self.selection_changed_confirm()
+        if changed:
+            self.selection_changed_confirm()
 
     def make_selection_valid(self):
         """ Make the selection valid and return the changed positions. """
@@ -1393,8 +1531,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
             self.plot.scene().removeItem(v)
         self.connected_views = []
 
-    def add_curves(self, x, ys, addc=True):
-        """ Add multiple curves with the same x domain. """
+    def _compute_sample(self, ys):
         if len(ys) > MAX_INSTANCES_DRAWN:
             sample_selection = \
                 random.Random(self.sample_seed).sample(range(len(ys)), MAX_INSTANCES_DRAWN)
@@ -1405,46 +1542,11 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
             subset_additional = MAX_INSTANCES_DRAWN - (len(subset) - len(subset_to_show))
             if len(subset_to_show) > subset_additional:
                 subset_to_show = \
-                    random.Random(self.sample_seed).sample(subset_to_show, subset_additional)
+                    random.Random(self.sample_seed).sample(sorted(subset_to_show), subset_additional)
             sampled_indices = sorted(sample_selection + list(subset_to_show))
-            self.sampling = True
         else:
             sampled_indices = list(range(len(ys)))
-        ys = self.data.X[sampled_indices][:, self.data_xsind]
-        ys[np.isinf(ys)] = np.nan  # remove infs that could ruin display
-
-        if self.waterfall:
-            waterfall_constant = 0.1
-            miny = bottleneck.nanmin(ys)
-            maxy = bottleneck.nanmax(ys)
-            space = (maxy - miny) * waterfall_constant
-            mul = (np.arange(len(ys))*space + 1).reshape(-1, 1)
-            ys = ys * mul
-
-        # shuffle the data before drawing because classes often appear sequentially
-        # and the last class would then seem the most prevalent if colored
-        indices = list(range(len(sampled_indices)))
-        random.Random(self.sample_seed).shuffle(indices)
-        sampled_indices = [sampled_indices[i] for i in indices]
-        self.sampled_indices = sampled_indices
-        self.sampled_indices_inverse = {s: i for i, s in enumerate(self.sampled_indices)}
-        self.new_sampling.emit(len(self.sampled_indices))
-        ys = ys[indices]  # ys was already subsampled
-
-        self.curves.append((x, ys))
-
-        # add curves efficiently
-        for y in ys:
-            self.add_curve(x, y, ignore_bounds=True)
-
-        if x.size and ys.size:
-            bounding_rect = QGraphicsRectItem(QRectF(
-                QPointF(bottleneck.nanmin(x), bottleneck.nanmin(ys)),
-                QPointF(bottleneck.nanmax(x), bottleneck.nanmax(ys))))
-            bounding_rect.setPen(QPen(Qt.NoPen))  # prevents border of 1
-            self.curves_cont.add_bounds(bounding_rect)
-
-        self.curves_plotted.append((x, ys))
+        return sampled_indices
 
     def add_curve(self, x, y, pen=None, ignore_bounds=False):
         c = FinitePlotCurveItem(x=x, y=y, pen=pen if pen else self.pen_normal[None])
@@ -1506,16 +1608,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
         self.legend.setVisible(bool(self.legend.items))
 
     def show_individual(self):
-        self.clear_graph()
-        self.view_average_menu.setChecked(False)
-        self.set_pen_colors()
-        self.viewtype = INDIVIDUAL
-        if not self.data:
-            return
-        self.add_curves(self.data_x, self.data.X)
-        self.set_curve_pens()
-        self.curves_cont.update()
-        self.plot.vb.set_mode_panning()
+        self.show_individual_thread.show()
 
     def resample_curves(self, seed):
         self.sample_seed = seed
@@ -1560,7 +1653,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
         self.view_waterfall_menu.setChecked(self.waterfall)
 
     def show_average(self):
-        self.show_average_thread.show_average()
+        self.show_average_thread.show()
 
     def update_view(self):
         if self.viewtype == INDIVIDUAL:
@@ -1644,6 +1737,7 @@ class CurvePlot(QWidget, OWComponent, SelectionGroupMixin):
 
     def shutdown(self):
         self.show_average_thread.shutdown()
+        self.show_individual_thread.shutdown()
 
     @classmethod
     def migrate_settings_sub(cls, settings, version):
